#include "../config.h"
#include <sys/types.h>
#include <assert.h>
#include <limits.h>
#include <pwd.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <dbus/dbus.h>
#include "common.h"
#include "mainloop.h"
#include "oddjob_dbus.h"
#include "util.h"

struct oddjob_dbus_context {
	DBusConnection *conn;
	char *name;
	struct oddjob_dbus_object {
		struct oddjob_dbus_context *ctx;
		char *path;
		struct oddjob_dbus_interface {
			char *interface;
			struct oddjob_dbus_method {
				char *method;
				int n_arguments;
				oddjob_dbus_handler *handler;
				void *data;
			} *methods;
			int n_methods;
		} *interfaces;
		int n_interfaces;
		dbus_bool_t registered;
	} *objects;
	int n_objects;
};

struct oddjob_dbus_message {
	DBusConnection *conn;
	DBusMessage *msg;
	int n_args;
	char **args;
};

static int
oddjob_dbus_bind(DBusConnection *conn, const char *service_name)
{
#if DBUS_CHECK_VERSION(0,30)
        return dbus_bus_request_name(conn, service_name, 0, NULL);
#elif DBUS_CHECK_VERSION(0,20)
	return dbus_bus_acquire_service(conn, service_name, 0, NULL);
#else
#error "Don't know how to set service names for your version of D-BUS!"
	return FALSE;
#endif
}

struct oddjob_dbus_context *
oddjob_dbus_listener_new(const char *service_name)
{
	DBusError err;
	struct oddjob_dbus_context *ctx;

	ctx = oddjob_malloc0(sizeof(struct oddjob_dbus_context));
	if (ctx == NULL) {
		return NULL;
	}

	memset(&err, 0, sizeof(err));
	ctx->conn = dbus_bus_get(DBUS_BUS_SYSTEM, &err);
	dbus_connection_ref(ctx->conn);

	/* Request all messages which we're allowed to see. */
	dbus_bus_add_match(ctx->conn, "", &err);

	/* Try to set ourselves up with a well-known service name.  If it
	 * fails, there may already be another copy running, but that's not our
	 * problem to solve. */
	if (!oddjob_dbus_bind(ctx->conn, service_name)) {
		fprintf(stderr, "Error binding to service name!\n");
		dbus_connection_unref(ctx->conn);
		oddjob_free(ctx);
		return NULL;
	}

	ctx->name = oddjob_strdup(service_name);
	ctx->n_objects = 0;
	ctx->objects = NULL;

	return ctx;
}

void
oddjob_dbus_listener_free(struct oddjob_dbus_context *ctx)
{
	int i, j, k;
	if (ctx == NULL) {
		return;
	}
	for (i = 0; i < ctx->n_objects; i++) {
		/* Clean up this object. */
		for (j = 0; j < ctx->objects[i].n_interfaces; j++) {
			/* Clean up this interface. */
			for (k = 0;
			     k < ctx->objects[i].interfaces[j].n_methods;
			     k++) {
				/* Clean up this method. */
				oddjob_free(ctx->objects[i].interfaces[j].methods[k].method);
				ctx->objects[i].interfaces[j].methods[k].method = NULL;
				ctx->objects[i].interfaces[j].methods[k].n_arguments = 0;
				ctx->objects[i].interfaces[j].methods[k].handler = NULL;
				ctx->objects[i].interfaces[j].methods[k].data = NULL;
			}
			oddjob_free(ctx->objects[i].interfaces[j].methods);
			ctx->objects[i].interfaces[j].methods = NULL;
			oddjob_free(ctx->objects[i].interfaces[j].interface);
			ctx->objects[i].interfaces[j].interface = NULL;
		}
		/* Clean up this object. */
		dbus_connection_unregister_object_path(ctx->conn,
						       ctx->objects[i].path);
		oddjob_free(ctx->objects[i].interfaces);
		ctx->objects[i].interfaces = NULL;
		oddjob_free(ctx->objects[i].path);
		ctx->objects[i].ctx = NULL;
	}
	/* Clean up this service. */
	oddjob_free(ctx->name);
	ctx->name = NULL;
	oddjob_free(ctx->objects);
	ctx->objects = NULL;
	ctx->n_objects = 0;
	dbus_connection_unref(ctx->conn);
	ctx->conn = NULL;
	oddjob_free(ctx);
}

static struct oddjob_dbus_message *
oddjob_dbus_message_from_message(DBusConnection *conn,
				 DBusMessage *message)
{
	struct oddjob_dbus_message *msg;
	char *p;
	dbus_bool_t more;
	DBusMessageIter iter;

	msg = oddjob_malloc0(sizeof(struct oddjob_dbus_message));
	msg->conn = conn;
	dbus_connection_ref(msg->conn);
	msg->msg = message;
	if (msg->msg != NULL) {
		dbus_message_ref(msg->msg);
		if (dbus_message_iter_init(message, &iter)) {
			more = TRUE;
			while (more) {
				switch (dbus_message_iter_get_arg_type(&iter)) {
				case DBUS_TYPE_STRING:
					oddjob_resize_array((void**) &msg->args,
							    sizeof(char*),
							    msg->n_args,
							    msg->n_args + 1);
#if DBUS_CHECK_VERSION(0,30)
					dbus_message_iter_get_basic(&iter, &p);
					msg->args[msg->n_args] = oddjob_strdup(p);
#elif DBUS_CHECK_VERSION(0,20)
					p = dbus_message_iter_get_string(&iter);
					msg->args[msg->n_args] = oddjob_strdup(p);
#else
#error "Don't know how to retrieve message arguments with your version of D-BUS!"
#endif
					msg->n_args++;
					break;
				case DBUS_TYPE_INVALID:
					more = FALSE;
					break;
				default:
					break;
				}
				if (!dbus_message_iter_has_next(&iter) ||
				    !dbus_message_iter_next(&iter)) {
					more = FALSE;
				}
			}
		}
	}
	return msg;
}

static void
oddjob_dbus_message_free(struct oddjob_dbus_message *msg)
{
	int i;
	if (msg != NULL) {
		if (msg->args != NULL) {
			for (i = 0; i < msg->n_args; i++) {
				oddjob_free(msg->args[i]);
			}
			oddjob_free(msg->args);
		}
		msg->args = NULL;
		msg->n_args = 0;
		if (msg->msg != NULL) {
			dbus_message_unref(msg->msg);
			msg->msg = NULL;
		}
		if (msg->conn != NULL) {
			dbus_connection_unref(msg->conn);
			msg->conn = NULL;
		}
		oddjob_free(msg);
	}
}

static DBusHandlerResult
oddjob_dbus_path_message_function(DBusConnection *conn,
				  DBusMessage *message,
				  void *user_data)
{
	DBusHandlerResult ret;
	struct oddjob_dbus_context *ctx;
	struct oddjob_dbus_object *obj;
	struct oddjob_dbus_interface *interface;
	struct oddjob_dbus_message *msg;
	char n_args[LINE_MAX];
	const char *client_bus_name, *called_path, *called_interface;
	unsigned long uid;
	struct passwd *pwd;
	int i;

	ctx = user_data;
	msg = NULL;
	ret = DBUS_HANDLER_RESULT_NOT_YET_HANDLED;

	/* Find the bus address of the message sender. */
	client_bus_name = dbus_message_get_sender(message);
	if (client_bus_name == NULL) {
		oddjob_dbus_send_message_response_error(msg,
							ODDJOB_INTERFACE_NAME
							".error"
							".unknown_sender_address",
							n_args);
		return DBUS_HANDLER_RESULT_HANDLED;
	}

	/* Get the called object path. */
	called_path = dbus_message_get_path(message);
	for (i = 0; i < ctx->n_objects; i++) {
		if ((called_path != NULL) &&
		    (strcmp(called_path, ctx->objects[i].path) == 0)) {
			break;
		}
	}
	if (i >= ctx->n_objects) {
		oddjob_dbus_send_message_response_error(msg,
							ODDJOB_INTERFACE_NAME
							".error"
							".object_not_present",
							called_path ?
							called_path : "");
		return DBUS_HANDLER_RESULT_HANDLED;
	}
	obj = &ctx->objects[i];

	/* Get the called interface. */
	called_interface = dbus_message_get_interface(message);
	for (i = 0; i < obj->n_interfaces; i++) {
		if ((called_interface != NULL) &&
		    (strcmp(called_interface,
			    obj->interfaces[i].interface) == 0)) {
			break;
		}
	}
	if (i >= obj->n_interfaces) {
		oddjob_dbus_send_message_response_error(msg,
							ODDJOB_INTERFACE_NAME
							".error"
							".interface_not_implemented",
							called_interface ?
							called_interface : "");
		return DBUS_HANDLER_RESULT_HANDLED;
	}
	interface = &obj->interfaces[i];

	/* Get the UID of the sending user and resolve it to a name. */
	uid = dbus_bus_get_unix_user(conn, client_bus_name, NULL);
	pwd = getpwuid(uid);
	if ((pwd == NULL) || (pwd->pw_uid != uid)) {
		snprintf(n_args, sizeof(n_args), "UID=%lu", uid);
		oddjob_dbus_send_message_response_error(msg,
							ODDJOB_INTERFACE_NAME
							".error"
							".unknown_user",
							n_args);
		return DBUS_HANDLER_RESULT_HANDLED;
	}

	/* Find the requested method. */
	for (i = 0; i < interface->n_methods; i++) {
		if (dbus_message_is_method_call(message,
						interface->interface,
						interface->methods[i].method)) {
			if (msg == NULL) {
				msg = oddjob_dbus_message_from_message(conn,
								       message);
			}
			if (msg->n_args != interface->methods[i].n_arguments) {
				snprintf(n_args, sizeof(n_args),
					 "wrong number of arguments: "
					 "expected %d, called with %d",
					 interface->methods[i].n_arguments,
					 msg->n_args);
				oddjob_dbus_send_message_response_error(msg,
									ODDJOB_INTERFACE_NAME
									".error"
									".invalid_call",
									n_args);
				ret = DBUS_HANDLER_RESULT_HANDLED;
				break;
			}
			interface->methods[i].handler(obj->ctx,
						      msg,
						      pwd->pw_name,
						      uid,
						      interface->methods[i].data);
			ret = DBUS_HANDLER_RESULT_HANDLED;
			break;
		}
	}
	if (msg != NULL) {
		oddjob_dbus_message_free(msg);
		msg = NULL;
	}
	return ret;
}

static struct DBusObjectPathVTable oddjob_vtable = {
	.unregister_function = NULL,
	.message_function = oddjob_dbus_path_message_function,
};

void
oddjob_dbus_listener_add_method(struct oddjob_dbus_context *ctx,
				const char *object_path,
				const char *interface,
				const char *method,
				int n_arguments,
				oddjob_dbus_handler *handler,
				void *data)
{
	int i;
	struct oddjob_dbus_object *obj;
	struct oddjob_dbus_interface *interf;
	struct oddjob_dbus_method *meth;

	/* find the object, creating it if it doesn't already exist */
	for (i = 0; i < ctx->n_objects; i++) {
		if ((ctx->objects[i].path != NULL) &&
		    (strcmp(object_path, ctx->objects[i].path) == 0)) {
			break;
		}
	}
	if (i >= ctx->n_objects) {
		oddjob_resize_array((void**)&ctx->objects,
				    sizeof(ctx->objects[0]),
				    ctx->n_objects, i + 1);
		ctx->objects[i].ctx = ctx;
		ctx->objects[i].path = oddjob_strdup(object_path);
		ctx->objects[i].interfaces = NULL;
		ctx->objects[i].n_interfaces = 0;
		ctx->objects[i].registered = FALSE;
		ctx->n_objects = i + 1;
	}
	obj = &ctx->objects[i];

	/* find the interface, creating it if it doesn't already exist */
	for (i = 0; i < obj->n_interfaces; i++) {
		if ((obj->interfaces[i].interface != NULL) &&
		    (strcmp(interface, obj->interfaces[i].interface) == 0)) {
			break;
		}
	}
	if (i >= obj->n_interfaces) {
		oddjob_resize_array((void**)&obj->interfaces,
				    sizeof(obj->interfaces[0]),
				    obj->n_interfaces, i + 1);
		obj->interfaces[i].interface = oddjob_strdup(interface);
		obj->n_interfaces = i + 1;
	}
	interf = &obj->interfaces[i];

	/* find the method, creating it if it doesn't already exist */
	for (i = 0; i < interf->n_methods; i++) {
		if ((interf->methods[i].method != NULL) &&
		    (strcmp(method, interf->methods[i].method) == 0)) {
			break;
		}
	}
	if (i >= interf->n_methods) {
		oddjob_resize_array((void**)&interf->methods,
				    sizeof(interf->methods[0]),
				    interf->n_methods, i + 1);
		interf->methods[i].method = oddjob_strdup(method);
		interf->n_methods = i + 1;
	}

	/* set the method's pointers */
	meth = &interf->methods[i];
	meth->n_arguments = n_arguments;
	meth->handler = handler;
	meth->data = data;

	/* last step - register any not-yet-registered objects */
	for (i = 0; i < ctx->n_objects; i++) {
		if ((!ctx->objects[i].registered) &&
		    (ctx->objects[i].path != NULL) &&
		    (strlen(ctx->objects[i].path) > 0)) {
			if (!dbus_connection_register_object_path(ctx->conn,
								  ctx->objects[i].path,
								  &oddjob_vtable,
								  ctx)) {
				fprintf(stderr, "Error registering object %s\n",
					ctx->objects[i].path);
				_exit(1);
			}
			ctx->objects[i].registered = TRUE;
		}
	}
}

const char *
oddjob_dbus_message_get_arg(struct oddjob_dbus_message *msg, int n)
{
	if (n >= msg->n_args) {
		return NULL;
	}
	return msg->args[n];
}

void
oddjob_dbus_send_message_response_success(struct oddjob_dbus_message *msg,
					  const char *text)
{
	DBusMessage *message;
	message = dbus_message_new_method_return(msg->msg);
	if ((text != NULL) && (strlen(text) > 0)) {
#if DBUS_CHECK_VERSION(0,30)
		dbus_message_append_args(message,
					 DBUS_TYPE_STRING, &text,
					 DBUS_TYPE_INVALID);
#elif DBUS_CHECK_VERSION(0,20)
		dbus_message_append_args(message,
					 DBUS_TYPE_STRING, text,
					 DBUS_TYPE_INVALID);
#else
#error		"Don't know how to set message arguments with your version of D-BUS!"
#endif
	}
	dbus_connection_send(msg->conn, message, NULL);
	dbus_message_unref(message);
}

void
oddjob_dbus_send_message_response_error(struct oddjob_dbus_message *msg,
				        const char *error,
				        const char *text)
{
	DBusMessage *message;
	message = dbus_message_new_error(msg->msg, error, text);
	dbus_connection_send(msg->conn, message, NULL);
	dbus_message_unref(message);
}

int
oddjob_dbus_call_method(const char *service, const char *object_path,
			const char *interface, const char *method,
			char *output, size_t output_length,
			...)
{
	DBusConnection *conn;
	DBusMessage *message, *reply;
	DBusError err;
	struct oddjob_dbus_message *msg;
	int ret;
	va_list ap;
	const char *p;

	if (output != NULL) {
		memset(output, '\0', output_length);
	}

	memset(&err, 0, sizeof(err));
	dbus_error_init(&err);
	conn = dbus_bus_get(DBUS_BUS_SYSTEM, &err);
	if (conn == NULL) {
		if (dbus_error_is_set(&err) && (output != NULL)) {
			snprintf(output, output_length, "%s: %s",
				 err.name, err.message);
		}
		dbus_error_free(&err);
		return -2;
	}

	dbus_connection_ref(conn);
	message = dbus_message_new_method_call(service,
					       object_path,
					       interface,
					       method);
	va_start(ap, output_length);
	while ((p = va_arg(ap, const char*)) != NULL) {
#if DBUS_CHECK_VERSION(0,30)
		dbus_message_append_args(message,
					 DBUS_TYPE_STRING, &p,
					 DBUS_TYPE_INVALID);
#elif DBUS_CHECK_VERSION(0,20)
		dbus_message_append_args(message,
					 DBUS_TYPE_STRING, p,
					 DBUS_TYPE_INVALID);
#else
#error		"Don't know how to set message arguments with your version of D-BUS!"
#endif
	}
	va_end(ap);
	reply = dbus_connection_send_with_reply_and_block(conn, message,
							  -1, &err);
	msg = oddjob_dbus_message_from_message(conn, reply);
	if ((output_length > 0) && (msg->n_args > 0)) {
		memset(output, '\0', output_length);
		strncpy(output, msg->args[0], output_length - 1);
	}

	if (dbus_error_is_set(&err)) {
		if (output != NULL) {
			snprintf(output, output_length, "%s: %s",
				 err.name, err.message);
		}
		dbus_error_free(&err);
		ret = -1;
	} else {
		ret = 0;
	}

	oddjob_dbus_message_free(msg);

	if (reply != NULL) {
		dbus_message_unref(reply);
	}

	dbus_message_unref(message);
	dbus_connection_unref(conn);

	return ret;
}

void
oddjob_dbus_main_init(struct oddjob_dbus_context *ctx)
{
	mainloop_reset_signal_handlers();
	mainloop_init(ctx->conn);
}

void
oddjob_dbus_main_done(struct oddjob_dbus_context *ctx)
{
	mainloop_done(ctx->conn);
}

int
oddjob_dbus_main_iterate(struct oddjob_dbus_context *ctx)
{
	return mainloop_iterate(ctx->conn);
}
